LIBRARY gumbo.dll
EXPORTS
clear_stack_to_table_body_context @1
clone_element_node @2
clone_node @3
consume_char_ref @4
get_appropriate_insertion_location @5
gumbo_add_error @6
gumbo_append_node @7
gumbo_attribute_set_value @8
gumbo_caret_diagnostic_to_string @9
gumbo_create_element_node @10
gumbo_create_node @11
gumbo_create_template_node @12
gumbo_create_text_node @13
gumbo_debug @14
gumbo_destroy_attribute @15
gumbo_destroy_errors @16
gumbo_destroy_node @17
gumbo_destroy_output @18
gumbo_element_remove_attribute @19
gumbo_element_remove_attribute_at @20
gumbo_element_set_attribute @21
gumbo_error_destroy @22
gumbo_error_to_string @23
gumbo_get_attribute @24
gumbo_init_errors @25
gumbo_insert_node @26
gumbo_lex @27
gumbo_memory_set_allocator @28
gumbo_memory_set_free @29
gumbo_new_output_init @30
gumbo_normalize_svg_tagname @31
gumbo_normalized_tagname @32
gumbo_parse @33
gumbo_parse_fragment @34
gumbo_parse_with_options @35
gumbo_print_caret_diagnostic @36
gumbo_remove_from_parent @37
gumbo_string_buffer_append_codepoint @38
gumbo_string_buffer_append_string @39
gumbo_string_buffer_clear @40
gumbo_string_buffer_cstr @41
gumbo_string_buffer_destroy @42
gumbo_string_buffer_init @43
gumbo_string_buffer_put @44
gumbo_string_buffer_putv @45
gumbo_string_buffer_reserve @46
gumbo_string_buffer_to_string @47
gumbo_string_copy @48
gumbo_string_equals @49
gumbo_string_equals_ignore_case @50
gumbo_tag_enum @51
gumbo_tag_from_original_text @52
gumbo_tagn_enum @53
gumbo_token_destroy @54
gumbo_tokenizer_set_is_current_node_foreign @55
gumbo_tokenizer_set_state @56
gumbo_tokenizer_state_destroy @57
gumbo_tokenizer_state_init @58
gumbo_user_allocator @59
gumbo_user_free @60
gumbo_vector_add @61
gumbo_vector_destroy @62
gumbo_vector_index_of @63
gumbo_vector_init @64
gumbo_vector_insert_at @65
gumbo_vector_pop @66
gumbo_vector_remove @67
gumbo_vector_remove_at @68
gumbo_vector_splice @69
kGumboDefaultOptions @70
kGumboEmptySourcePosition @71
kGumboEmptyString @72
kGumboEmptyVector @73
kGumboNoChar @74
kGumboTagNames @75
kScriptTag @76
kUtf8ReplacementChar @77
utf8_is_invalid_code_point @78
utf8iterator_current @79
utf8iterator_fill_error_at_mark @80
utf8iterator_get_char_pointer @81
utf8iterator_get_end_pointer @82
utf8iterator_get_position @83
utf8iterator_init @84
utf8iterator_mark @85
utf8iterator_maybe_consume_match @86
utf8iterator_next @87
utf8iterator_reset @88
